module LogicUS.FOL.SintaxSemantics exposing
    ( Term, FormulaFOL, SetFOL, termVarSymbols, termsVarSymbols, ffolVarSymbols, ffolFormTree, ffolIsWellFormed, ffolHasInstanceOfVar, ffolHasFreeInstanceOfVar, ffolHasLinkedInstanceOfVar, termIsClosed, ffolIsOpen, ffolIsClosed
    , Substitution, substitutionDomain, termApplySubstitution, ffolApplySubstitution, substitutionComposition, renameVars, ffolUniversalClausure, ffolExistencialClausure
    , L_Structure, termInterpretation, termsInterpretation, ffolInterpretation, sfolInterpretation
    , ffolReadFromString, ffolReadExtraction, ffolToInputString, substitutionReadFromString, substitutionReadExtraction, substitutionToInputString
    , ffolToString, sfolToString, ffolToMathString, sfolToMathString, formTreeToString, formTreeToDOT, substitutionToString, substitutionToMathString, l_StructureToString
    )

{-| The module provides the elementary tools for working with first order logic. It allows defining both formulas and sets as well as performing some basic operations on them, such as evaluations regarding interpretations, performs substitutions, clausures, ...


# FOL Formulas and Sets

@docs Term, FormulaFOL, SetFOL, termVarSymbols, termsVarSymbols, ffolVarSymbols, ffolFormTree, ffolIsWellFormed, ffolHasInstanceOfVar, ffolHasFreeInstanceOfVar, ffolHasLinkedInstanceOfVar, termIsClosed, ffolIsOpen, ffolIsClosed


# Substitutions, variable rename and clausure

@docs Substitution, substitutionDomain, termApplySubstitution, ffolApplySubstitution, substitutionComposition, renameVars, ffolUniversalClausure, ffolExistencialClausure


# L-structures and valuations

@docs L_Structure, termInterpretation, termsInterpretation, ffolInterpretation, sfolInterpretation


# Parsers

@docs ffolReadFromString, ffolReadExtraction, ffolToInputString, substitutionReadFromString, substitutionReadExtraction, substitutionToInputString


# Representation

@docs ffolToString, sfolToString, ffolToMathString, sfolToMathString, formTreeToString, formTreeToDOT, substitutionToString, substitutionToMathString, l_StructureToString

-}

--===========--
--  IMPORTS  --
--===========--

import Dict exposing (Dict)
import Graph exposing (Edge, Graph, Node, NodeId)
import Graph.DOT exposing (defaultStyles)
import List.Extra as LE
import Maybe.Extra as ME
import Parser exposing ((|.), (|=), Parser, Trailing(..), succeed)
import String.Extra as SE



-----------
-- TYPES --
-----------


{-| It is used to represent the terms in First Order Logic , these are variables, constants and functions. Note that constants are a particular case of functions, which are not dependent on any variable.
-}
type Term
    = Var String
    | Func String (List Term)


{-| It is a recursive definition of a formula in First Order Logic. It could be a predicate, equality, universal cuantification, existencial cuantificacation, negation, conjuction, implication, equivalence and unsatisfiable formula
-}
type FormulaFOL
    = Pred String (List Term)
    | Equal Term Term
    | Neg FormulaFOL
    | Conj FormulaFOL FormulaFOL
    | Disj FormulaFOL FormulaFOL
    | Impl FormulaFOL FormulaFOL
    | Equi FormulaFOL FormulaFOL
    | Exists Term FormulaFOL
    | Forall Term FormulaFOL
    | Insat


{-| It is used to define sets of formulas in First Order Logic.
-}
type alias SetFOL =
    List FormulaFOL



--===============--
--  FUNCTIONS    --
--===============--
----------------------
-- Regular functions -
----------------------
-- {-| It decides if a term is a variable or not
-- -}
-- termIsVariable : Term -> Bool
-- termIsVariable t =
--     case t of
--         Var _ ->
--             True
--         _ ->
--             False
-- {-| It decides if a term is a constant or not
-- -}
-- termIsConstant : Term -> Bool
-- termIsConstant t =
--     case t of
--         Func _ [] ->
--             True
--         _ ->
--             False
-- {-| It gets the symbol of a variable. If the argument is not a variable it returns Nothing.
-- -}
-- termGetVarSymb : Term -> Maybe String
-- termGetVarSymb t =
--     case t of
--         Var x ->
--             Just x
--         _ ->
--             Nothing
-- {-| It gets the symbol of a constant. If the argument is not a constant it returns Nothing.
-- -}
-- termGetConstSymb : Term -> Maybe String
-- termGetConstSymb t =
--     case t of
--         Func x [] ->
--             Just x
--         _ ->
--             Nothing


{-| It gets all the variable symbols that acts inside a term

    termVarSymbols (Func "f" [ Var "a", Func "c" [], Func "g" [ Var "b", Var "a" ] ]) =
        [ "a", "b" ]

-}
termVarSymbols : Term -> List String
termVarSymbols t =
    case t of
        Var x ->
            [ x ]

        Func _ ts ->
            List.sort <| LE.unique <| List.concat <| List.map termVarSymbols ts


{-| It gets all the variables that acts inside a list of terms

    termsVarSymbols [ Func "f" [ Var "a", Func "c" [] ], Func "g" [ Var "b", Var "c" ] ] =
        [ "a", "b", "c" ]

-}
termsVarSymbols : List Term -> List String
termsVarSymbols ts =
    List.sort <| LE.unique <| List.concat <| List.map termVarSymbols ts


{-| It gets all the variables that acts inside a formula

    f1 =
        ffolReadExtraction <| ffolReadFromString "_p(x,y)=_p(y,x)"

    f2 =
        ffolReadExtraction <| ffolReadFromString "!A[x] !A[y] (M(x,y) -> !E[z] F(z, _jhon) & P(x,z) | P(y,z))"

    ffolVarSymbols f1 =
        [ "x", "y" ]
    ffolVarSymbols f2 =
        [ "x", "y", "z" ]

-}
ffolVarSymbols : FormulaFOL -> List String
ffolVarSymbols f =
    case f of
        Pred _ ts ->
            termsVarSymbols ts

        Equal t1 t2 ->
            termsVarSymbols [ t1, t2 ]

        Neg p ->
            ffolVarSymbols p

        Conj p q ->
            List.sort <| LE.unique <| ffolVarSymbols p ++ ffolVarSymbols q

        Disj p q ->
            List.sort <| LE.unique <| ffolVarSymbols p ++ ffolVarSymbols q

        Impl p q ->
            List.sort <| LE.unique <| ffolVarSymbols p ++ ffolVarSymbols q

        Equi p q ->
            List.sort <| LE.unique <| ffolVarSymbols p ++ ffolVarSymbols q

        Exists (Var x) p ->
            List.sort <| LE.unique <| ffolVarSymbols p ++ [ x ]

        Forall (Var x) p ->
            List.sort <| LE.unique <| ffolVarSymbols p ++ [ x ]

        _ ->
            []


{-| It gives a graph with the form tree of a formula. If you want visualize it you can use formTreeToDOT and visualize the result witn an online graphviz render. Don't forget change `\\n` by `\n` and `\\"` by `"` in a code editor previously.

    ffolFormTree f2 |> formTreeToDOT =
        "digraph G {\n  rankdir=TB\n  graph []\n  node [shape=plaintext, color=black]\n  edge [dir=none]\n\n  0 -> 1\n  1 -> 2\n  2 -> 3\n  2 -> 9\n  3 -> 4\n  3 -> 5\n  5 -> 6\n  5 -> 8\n  6 -> 7\n\n  0 [label=\"∀x ∀y ( ( M(x, y) → ( ∃z F(z, jhon) ∧ P(x, z) ) ) ∨ P(y, z) )\"]\n  1 [label=\"∀y ( ( M(x, y) → ( ∃z F(z, jhon) ∧ P(x, z) ) ) ∨ P(y, z) )\"]\n  2 [label=\"( ( M(x, y) → ( ∃z F(z, jhon) ∧ P(x, z) ) ) ∨ P(y, z) )\"]\n  3 [label=\"( M(x, y) → ( ∃z F(z, jhon) ∧ P(x, z) ) )\"]\n  4 [label=\"M(x, y)\"]\n  5 [label=\"( ∃z F(z, jhon) ∧ P(x, z) )\"]\n  6 [label=\"∃z F(z, jhon)\"]\n  7 [label=\"F(z, jhon)\"]\n  8 [label=\"P(x, z)\"]\n  9 [label=\"P(y, z)\"]\n}"

Visualize the [example](https://dreampuf.github.io/GraphvizOnline/#digraph%20G%20%7B%0A%20%20rankdir%3DTB%0A%20%20graph%20%5B%5D%0A%20%20node%20%5Bshape%3Dplaintext%2C%20color%3Dblack%5D%0A%20%20edge%20%5Bdir%3Dnone%5D%0A%0A%20%200%20-%3E%201%0A%20%201%20-%3E%202%0A%20%202%20-%3E%203%0A%20%202%20-%3E%209%0A%20%203%20-%3E%204%0A%20%203%20-%3E%205%0A%20%205%20-%3E%206%0A%20%205%20-%3E%208%0A%20%206%20-%3E%207%0A%0A%20%200%20%5Blabel%3D%22%E2%88%80x%20%E2%88%80y%20(%20(%20M(x%2C%20y)%20%E2%86%92%20(%20%E2%88%83z%20F(z%2C%20jhon)%20%E2%88%A7%20P(x%2C%20z)%20)%20)%20%E2%88%A8%20P(y%2C%20z)%20)%22%5D%0A%20%201%20%5Blabel%3D%22%E2%88%80y%20(%20(%20M(x%2C%20y)%20%E2%86%92%20(%20%E2%88%83z%20F(z%2C%20jhon)%20%E2%88%A7%20P(x%2C%20z)%20)%20)%20%E2%88%A8%20P(y%2C%20z)%20)%22%5D%0A%20%202%20%5Blabel%3D%22(%20(%20M(x%2C%20y)%20%E2%86%92%20(%20%E2%88%83z%20F(z%2C%20jhon)%20%E2%88%A7%20P(x%2C%20z)%20)%20)%20%E2%88%A8%20P(y%2C%20z)%20)%22%5D%0A%20%203%20%5Blabel%3D%22(%20M(x%2C%20y)%20%E2%86%92%20(%20%E2%88%83z%20F(z%2C%20jhon)%20%E2%88%A7%20P(x%2C%20z)%20)%20)%22%5D%0A%20%204%20%5Blabel%3D%22M(x%2C%20y)%22%5D%0A%20%205%20%5Blabel%3D%22(%20%E2%88%83z%20F(z%2C%20jhon)%20%E2%88%A7%20P(x%2C%20z)%20)%22%5D%0A%20%206%20%5Blabel%3D%22%E2%88%83z%20F(z%2C%20jhon)%22%5D%0A%20%207%20%5Blabel%3D%22F(z%2C%20jhon)%22%5D%0A%20%208%20%5Blabel%3D%22P(x%2C%20z)%22%5D%0A%20%209%20%5Blabel%3D%22P(y%2C%20z)%22%5D%0A%7D)

-}
ffolFormTree : FormulaFOL -> Graph FormulaFOL ()
ffolFormTree x =
    case x of
        Pred _ _ ->
            Graph.fromNodesAndEdges [ Node 0 x ] []

        Equal _ _ ->
            Graph.fromNodesAndEdges [ Node 0 x ] []

        Neg p ->
            let
                ( nodes, edges ) =
                    formTreeAux p 1
            in
            Graph.fromNodesAndEdges (Node 0 x :: nodes) (Edge 0 1 () :: edges)

        Conj p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p 1
            in
            let
                nextid =
                    List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            Graph.fromNodesAndEdges (Node 0 x :: (nodes1 ++ nodes2)) ([ Edge 0 1 (), Edge 0 nextid () ] ++ edges1 ++ edges2)

        Disj p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p 1
            in
            let
                nextid =
                    List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            Graph.fromNodesAndEdges (Node 0 x :: (nodes1 ++ nodes2)) ([ Edge 0 1 (), Edge 0 nextid () ] ++ edges1 ++ edges2)

        Impl p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p 1
            in
            let
                nextid =
                    List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            Graph.fromNodesAndEdges (Node 0 x :: (nodes1 ++ nodes2)) ([ Edge 0 1 (), Edge 0 nextid () ] ++ edges1 ++ edges2)

        Equi p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p 1
            in
            let
                nextid =
                    List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            Graph.fromNodesAndEdges (Node 0 x :: (nodes1 ++ nodes2)) ([ Edge 0 1 (), Edge 0 nextid () ] ++ edges1 ++ edges2)

        Exists _ p ->
            let
                ( nodes, edges ) =
                    formTreeAux p 1
            in
            Graph.fromNodesAndEdges (Node 0 x :: nodes) (Edge 0 1 () :: edges)

        Forall _ p ->
            let
                ( nodes, edges ) =
                    formTreeAux p 1
            in
            Graph.fromNodesAndEdges (Node 0 x :: nodes) (Edge 0 1 () :: edges)

        Insat ->
            Graph.fromNodesAndEdges [ Node 0 x ] []


formTreeAux : FormulaFOL -> NodeId -> ( List (Node FormulaFOL), List (Edge ()) )
formTreeAux x nodeid =
    case x of
        Pred _ _ ->
            ( [ Node nodeid x ], [] )

        Equal _ _ ->
            ( [ Node nodeid x ], [] )

        Neg p ->
            let
                ( nodes, edges ) =
                    formTreeAux p (nodeid + 1)
            in
            ( Node nodeid x :: nodes, Edge nodeid (nodeid + 1) () :: edges )

        Conj p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p (nodeid + 1)
            in
            let
                nextid =
                    nodeid + List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            ( Node nodeid x :: (nodes1 ++ nodes2), [ Edge nodeid (nodeid + 1) (), Edge nodeid nextid () ] ++ edges1 ++ edges2 )

        Disj p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p (nodeid + 1)
            in
            let
                nextid =
                    nodeid + List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            ( Node nodeid x :: (nodes1 ++ nodes2), [ Edge nodeid (nodeid + 1) (), Edge nodeid nextid () ] ++ edges1 ++ edges2 )

        Impl p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p (nodeid + 1)
            in
            let
                nextid =
                    nodeid + List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            ( Node nodeid x :: (nodes1 ++ nodes2), [ Edge nodeid (nodeid + 1) (), Edge nodeid nextid () ] ++ edges1 ++ edges2 )

        Equi p q ->
            let
                ( nodes1, edges1 ) =
                    formTreeAux p (nodeid + 1)
            in
            let
                nextid =
                    nodeid + List.length nodes1 + 1
            in
            let
                ( nodes2, edges2 ) =
                    formTreeAux q nextid
            in
            ( Node nodeid x :: (nodes1 ++ nodes2), [ Edge nodeid (nodeid + 1) (), Edge nodeid nextid () ] ++ edges1 ++ edges2 )

        Exists _ p ->
            let
                ( nodes, edges ) =
                    formTreeAux p (nodeid + 1)
            in
            ( Node nodeid x :: nodes, Edge nodeid (nodeid + 1) () :: edges )

        Forall _ p ->
            let
                ( nodes, edges ) =
                    formTreeAux p (nodeid + 1)
            in
            ( Node nodeid x :: nodes, Edge nodeid (nodeid + 1) () :: edges )

        Insat ->
            ( [ Node nodeid x ], [] )


{-| It indicates if a formula is well formed or not. A formula is well formed if it doesn't contain two nested cuantifiers over the same variable

    f3 = ffolReadExtraction <| ffolReadFromString "!A[x] !A[y] (M(x,y) -> !E[x] F(x, _jhon) & P(x,x) | P(y,x))"

    ffolIsWellFormed f1 == True
    ffolIsWellFormed f2 == True
    ffolIsWellFormed f3 == False

-}
ffolIsWellFormed : FormulaFOL -> Bool
ffolIsWellFormed f =
    let
        aux g ls =
            case g of
                Pred _ _ ->
                    True

                Equal _ _ ->
                    True

                Neg p ->
                    aux p ls

                Conj p q ->
                    aux p ls && aux q ls

                Disj p q ->
                    aux p ls && aux q ls

                Impl p q ->
                    aux p ls && aux q ls

                Equi p q ->
                    aux p ls && aux q ls

                Exists (Var x) p ->
                    not (List.member x ls) && aux p (ls ++ [ x ])

                Forall (Var x) p ->
                    not (List.member x ls) && aux p (ls ++ [ x ])

                Insat ->
                    True

                _ ->
                    False
    in
    aux f []


{-| It gives if a formula contains an instance of of one var with the symbol given.
-}
ffolHasInstanceOfVar : FormulaFOL -> String -> Bool
ffolHasInstanceOfVar f v =
    case f of
        Pred _ ts ->
            List.member v (termsVarSymbols ts)

        Equal t1 t2 ->
            List.member v (termsVarSymbols [ t1, t2 ])

        Neg g ->
            ffolHasInstanceOfVar g v

        Conj g h ->
            ffolHasInstanceOfVar g v || ffolHasInstanceOfVar h v

        Disj g h ->
            ffolHasInstanceOfVar g v || ffolHasInstanceOfVar h v

        Impl g h ->
            ffolHasInstanceOfVar g v || ffolHasInstanceOfVar h v

        Equi g h ->
            ffolHasInstanceOfVar g v || ffolHasInstanceOfVar h v

        Exists (Var _) g ->
            ffolHasInstanceOfVar g v

        Forall (Var _) g ->
            ffolHasInstanceOfVar g v

        _ ->
            False


{-| It checks if in one formula if it exists any free instance of a given variable

    f4 = ffolReadExtraction <| ffolReadFromString "P(x,y) -> !E[x] Q(x,x) "

    ffolHasFreeInstanceOfVar f1 "x" == True
    ffolHasFreeInstanceOfVar f2 "x" == False
    ffolHasFreeInstanceOfVar f4 "x" == True

-}
ffolHasFreeInstanceOfVar : FormulaFOL -> String -> Bool
ffolHasFreeInstanceOfVar f v =
    case f of
        Pred _ ts ->
            List.member v (termsVarSymbols ts)

        Equal t1 t2 ->
            List.member v (termsVarSymbols [ t1, t2 ])

        Neg g ->
            ffolHasFreeInstanceOfVar g v

        Conj g h ->
            ffolHasFreeInstanceOfVar g v || ffolHasFreeInstanceOfVar h v

        Disj g h ->
            ffolHasFreeInstanceOfVar g v || ffolHasFreeInstanceOfVar h v

        Impl g h ->
            ffolHasFreeInstanceOfVar g v || ffolHasFreeInstanceOfVar h v

        Equi g h ->
            ffolHasFreeInstanceOfVar g v || ffolHasFreeInstanceOfVar h v

        Exists (Var x) g ->
            not (x == v) && ffolHasFreeInstanceOfVar g v

        Forall (Var x) g ->
            not (x == v) && ffolHasFreeInstanceOfVar g v

        _ ->
            False


{-| It checks if in one formula if it exists any linked instance of a given variable

    ffolHasLinkedInstanceOfVar f1 "x" == False

    ffolHasLinkedInstanceOfVar f2 "x" == True

    ffolHasLinkedInstanceOfVar f4 "x" == True

-}
ffolHasLinkedInstanceOfVar : FormulaFOL -> String -> Bool
ffolHasLinkedInstanceOfVar f v =
    case f of
        Pred _ _ ->
            False

        Equal _ _ ->
            False

        Neg g ->
            ffolHasLinkedInstanceOfVar g v

        Conj g h ->
            ffolHasLinkedInstanceOfVar g v || ffolHasLinkedInstanceOfVar h v

        Disj g h ->
            ffolHasLinkedInstanceOfVar g v || ffolHasLinkedInstanceOfVar h v

        Impl g h ->
            ffolHasLinkedInstanceOfVar g v || ffolHasLinkedInstanceOfVar h v

        Equi g h ->
            ffolHasLinkedInstanceOfVar g v || ffolHasLinkedInstanceOfVar h v

        Exists (Var x) g ->
            (x == v) && ffolHasInstanceOfVar g v || ffolHasLinkedInstanceOfVar g v

        Forall (Var x) g ->
            (x == v) && ffolHasInstanceOfVar g v || ffolHasLinkedInstanceOfVar g v

        _ ->
            False


{-| It determinates if a term is closed this is if it doesn't contains any variable
-}
termIsClosed : Term -> Bool
termIsClosed t =
    (List.isEmpty << termVarSymbols) t


{-| It determinates if a formula is opened this is if it doesn't contains any cuantifier
-}
ffolIsOpen : FormulaFOL -> Bool
ffolIsOpen f =
    case f of
        Pred _ _ ->
            True

        Equal _ _ ->
            True

        Neg g ->
            ffolIsOpen g

        Conj g h ->
            ffolIsOpen g && ffolIsOpen h

        Disj g h ->
            ffolIsOpen g && ffolIsOpen h

        Impl g h ->
            ffolIsOpen g && ffolIsOpen h

        Equi g h ->
            ffolIsOpen g && ffolIsOpen h

        Exists _ _ ->
            False

        Forall _ _ ->
            False

        Insat ->
            True


{-| It checks if s formula is closed this is if it hasn't any free variables
-}
ffolIsClosed : FormulaFOL -> Bool
ffolIsClosed f =
    not <| List.any (\x -> ffolHasFreeInstanceOfVar f x) (ffolVarSymbols f)



-------------------
-- SUBSTITUTIONS --
-------------------


{-| It defines a substitution in First Order Logic
-}
type alias Substitution =
    Dict String Term


{-| It gets the variable domain of the substitution, that is variables symbols that participates in it.
-}
substitutionDomain : Substitution -> List String
substitutionDomain s =
    Dict.keys s


{-| It performes a substitution in a term, replacing the variables according to the substitution.
-}
termApplySubstitution : Substitution -> Term -> Term
termApplySubstitution s t =
    case t of
        Var x ->
            Maybe.withDefault (Var x) <| Dict.get x s

        Func f_ ts ->
            Func f_ <| List.map (termApplySubstitution s) ts


{-| It performes a substitution in a formula, replacing the variables according to the substitution.
-}
ffolApplySubstitution : Substitution -> FormulaFOL -> FormulaFOL
ffolApplySubstitution s f =
    case f of
        Pred p_ ts ->
            Pred p_ <| List.map (termApplySubstitution s) ts

        Equal t1 t2 ->
            Equal (termApplySubstitution s t1) (termApplySubstitution s t2)

        Neg g ->
            Neg (ffolApplySubstitution s g)

        Conj g h ->
            Conj (ffolApplySubstitution s g) (ffolApplySubstitution s h)

        Disj g h ->
            Disj (ffolApplySubstitution s g) (ffolApplySubstitution s h)

        Impl g h ->
            Impl (ffolApplySubstitution s g) (ffolApplySubstitution s h)

        Equi g h ->
            Equi (ffolApplySubstitution s g) (ffolApplySubstitution s h)

        Exists (Var x) g ->
            Exists (Var x) <| ffolApplySubstitution (Dict.filter (\k v -> k /= x && (not <| List.member x <| termVarSymbols v)) s) g

        Forall (Var x) g ->
            Forall (Var x) <| ffolApplySubstitution (Dict.filter (\k v -> k /= x && (not <| List.member x <| termVarSymbols v)) s) g

        _ ->
            Insat


{-| It gives the composition of two substitutions (s1 ∘ s2)
-}
substitutionComposition : Substitution -> Substitution -> Substitution
substitutionComposition s1 s2 =
    Dict.fromList <|
        List.filter (\( x, t ) -> Var x /= t) <|
            List.map (\( x2, t2 ) -> ( x2, termApplySubstitution s1 t2 )) (Dict.toList s2)
                ++ List.filter (\( x1, _ ) -> not (Dict.member x1 s2)) (Dict.toList s1)


{-| It renames variables in a formula if it is necessary. If a instance is linked to several cuantifiers (non well formed formula) it takes the nearest one as reference.
-}
renameVars : FormulaFOL -> FormulaFOL
renameVars f =
    let
        aux g cur hist =
            case g of
                Pred n ts ->
                    let
                        s =
                            Dict.map (\k v -> Var (k ++ "_{" ++ String.fromInt v ++ "}")) cur
                    in
                    ( Pred n <| List.map (\t -> termApplySubstitution s t) ts, hist )

                Equal t1 t2 ->
                    let
                        s =
                            Dict.map (\k v -> Var (k ++ "_{" ++ String.fromInt v ++ "}")) cur
                    in
                    ( Equal (termApplySubstitution s t1) (termApplySubstitution s t2), hist )

                Neg h ->
                    let
                        ( nh, nhist ) =
                            aux h cur hist
                    in
                    ( Neg nh, nhist )

                Conj h i ->
                    let
                        ( nh, nhist1 ) =
                            aux h cur hist
                    in
                    let
                        ( ni, nhist2 ) =
                            aux i cur nhist1
                    in
                    ( Conj nh ni, nhist2 )

                Disj h i ->
                    let
                        ( nh, nhist1 ) =
                            aux h cur hist
                    in
                    let
                        ( ni, nhist2 ) =
                            aux i cur nhist1
                    in
                    ( Disj nh ni, nhist2 )

                Impl h i ->
                    let
                        ( nh, nhist1 ) =
                            aux h cur hist
                    in
                    let
                        ( ni, nhist2 ) =
                            aux i cur nhist1
                    in
                    ( Impl nh ni, nhist2 )

                Equi h i ->
                    let
                        ( nh, nhist1 ) =
                            aux h cur hist
                    in
                    let
                        ( ni, nhist2 ) =
                            aux i cur nhist1
                    in
                    ( Equi nh ni, nhist2 )

                Forall (Var x) h ->
                    let
                        xind =
                            (Maybe.withDefault 0 <| Dict.get x cur) + 1
                    in
                    let
                        ncur =
                            Dict.insert x xind cur

                        nhist1 =
                            Dict.insert x xind hist
                    in
                    let
                        ( nh, nhist2 ) =
                            aux h ncur nhist1
                    in
                    ( Forall (Var (x ++ "_{" ++ String.fromInt xind ++ "}")) nh, nhist2 )

                Exists (Var x) h ->
                    let
                        xind =
                            (Maybe.withDefault 0 <| Dict.get x cur) + 1
                    in
                    let
                        ncur =
                            Dict.insert x xind cur

                        nhist1 =
                            Dict.insert x xind hist
                    in
                    let
                        ( nh, nhist2 ) =
                            aux h ncur nhist1
                    in
                    ( Exists (Var (x ++ "_{" ++ String.fromInt xind ++ "}")) nh, nhist2 )

                _ ->
                    ( Insat, hist )
    in
    Tuple.first <| aux f Dict.empty Dict.empty


{-| It gives the universal clausure of a formula
-}
ffolUniversalClausure : FormulaFOL -> FormulaFOL
ffolUniversalClausure f =
    let
        openVars =
            List.filter (\v -> ffolHasFreeInstanceOfVar f v) <| ffolVarSymbols f
    in
    List.foldl (\x ac -> Forall (Var x) ac) f openVars


{-| It gives the existencial clausure of a formula
-}
ffolExistencialClausure : FormulaFOL -> FormulaFOL
ffolExistencialClausure f =
    let
        openVars =
            List.filter (\v -> ffolHasFreeInstanceOfVar f v) <| ffolVarSymbols f
    in
    List.foldl (\x ac -> Exists (Var x) ac) f openVars



---------------------
-- INTERPRETATIONS --
---------------------


{-| It defines an interpretation in First Order Logic, this is a pair of a list of elements (universe) and a record with the following keys and values:

  - key:`const`, value: A dictionary that asigns to each constant symbol a element of the universe.

  - key:`funct`, value: A dictionary that asigns to each function symbol, $f$, a tuple whose first argument regards to the arity, $n$, and second one a function $U^{n} \\longrightarrow U$ , that takes a list with $n$ elements of the universe and returns a element of the universe.

  - key:`pred`, value: A dictionary that asigns to each predicate symbol, $P$, a tuple whose first argument regards to the arity, $n$, and second one a predicate $U^{n} \\longrightarrow {F, T}$ , that takes a list with $n$ elements of the universe and returns a boolean value.

-}
type alias L_Structure a =
    ( List a
    , { const : Dict String a
      , func : Dict String ( Int, List a -> Maybe a )
      , pred : Dict String ( Int, List a -> Maybe Bool )
      }
    )


{-| It calculates the interpretation of a term regarding to a L\_structure. If it is not interpretable by the L\_struncture then it returns Nothing.
-}
termInterpretation : Term -> L_Structure a -> Maybe a
termInterpretation t ( u, i ) =
    case t of
        Var _ ->
            Nothing

        Func f_ [] ->
            Dict.get f_ i.const

        Func f_ ts ->
            Maybe.andThen
                (\( ar, f_M ) ->
                    if List.length ts == ar then
                        Maybe.andThen
                            f_M
                            (termsInterpretation ts ( u, i ))

                    else
                        Nothing
                )
                (Dict.get f_ i.func)


{-| It calculates the interpretations of a list of terms, if any of them is not interpretable then it returns Nothing.
-}
termsInterpretation : List Term -> L_Structure a -> Maybe (List a)
termsInterpretation ts l =
    ME.combine <| List.map (\t -> termInterpretation t l) ts


{-| It calculates the valuation of a formula regarding to an interpretation. If the formula is not closed or any element is not interpretable it returns Nothing.
-}
ffolInterpretation : FormulaFOL -> L_Structure a -> Maybe Bool
ffolInterpretation f l =
    if (not << ffolIsClosed) f then
        Nothing

    else
        ffolInterpretationAux f l


ffolInterpretationAux : FormulaFOL -> L_Structure a -> Maybe Bool
ffolInterpretationAux f ( u, i ) =
    case f of
        Pred p_ ts ->
            Maybe.andThen
                (\( ar, p_M ) ->
                    if List.length ts == ar then
                        Maybe.andThen
                            p_M
                            (termsInterpretation ts ( u, i ))

                    else
                        Nothing
                )
                (Dict.get p_ i.pred)

        Equal t1 t2 ->
            Maybe.map2 (==) (termInterpretation t1 ( u, i )) (termInterpretation t2 ( u, i ))

        Neg g ->
            Maybe.map not (ffolInterpretation g ( u, i ))

        Conj g h ->
            Maybe.map2 (&&) (ffolInterpretation g ( u, i )) (ffolInterpretation h ( u, i ))

        Disj g h ->
            Maybe.map2 (||) (ffolInterpretation g ( u, i )) (ffolInterpretation h ( u, i ))

        Impl g h ->
            Maybe.map2 (\x y -> not x || y) (ffolInterpretation g ( u, i )) (ffolInterpretation h ( u, i ))

        Equi g h ->
            Maybe.map2 (==) (ffolInterpretation g ( u, i )) (ffolInterpretation h ( u, i ))

        Exists (Var x) g ->
            Maybe.map (List.any identity) <|
                ME.combine <|
                    List.map
                        (\o -> ffolInterpretationAux (ffolApplySubstitution (Dict.singleton x (Func o [])) g) ( u, i ))
                        (Dict.keys i.const)

        Forall (Var x) g ->
            Maybe.map (List.all identity) <|
                ME.combine <|
                    List.map
                        (\o -> ffolInterpretationAux (ffolApplySubstitution (Dict.singleton x (Func o [])) g) ( u, i ))
                        (Dict.keys i.const)

        _ ->
            Nothing


{-| It calculates the valuation of a set of formulas regarding to an interpretation. If any formula is interpretable it returns Nothing.
-}
sfolInterpretation : SetFOL -> L_Structure a -> Maybe Bool
sfolInterpretation fs l =
    Maybe.map (List.all identity) <|
        ME.combine <|
            List.map (\f -> ffolInterpretation f l) fs



------------
-- PARSER --
------------
-- It removes the spaces of a string


cleanSpaces : String -> String
cleanSpaces x =
    String.join "" <| String.split " " <| SE.clean x


{-| It reads the formula from a string. It returns a tuple with may be a formula (if it can be read it), the input considerated to parse and a message of error it it is not able to performs the parsing.

Messages are not perfect but we're working to improve it.

-}
ffolReadFromString : String -> ( Maybe FormulaFOL, String, String )
ffolReadFromString x =
    case cleanSpaces x of
        "" ->
            ( Maybe.Nothing, "", "Argument is empty" )

        s ->
            case Parser.run ffolParser ("(" ++ s ++ ")") of
                Ok y ->
                    ( Maybe.Just y, ffolToInputString y, "" )

                Err y ->
                    ( Maybe.Nothing, "(" ++ s ++ ")", "Error: " ++ String.replace "\"" "'" (Debug.toString y) )


{-| It extract the formula readed. If it is Nothing it returns Insat
-}
ffolReadExtraction : ( Maybe FormulaFOL, String, String ) -> FormulaFOL
ffolReadExtraction ( f, _, _ ) =
    Maybe.withDefault Insat f


termToInputString : Term -> String
termToInputString x =
    case x of
        Var s ->
            s

        Func n ts ->
            "_" ++ n ++ termsToInputString ts


termsToInputString : List Term -> String
termsToInputString ts =
    if List.isEmpty ts then
        ""

    else
        "(" ++ (String.join ", " <| List.map termToInputString ts) ++ ")"


{-| It gives the corresponding input syntax of a formula
-}
ffolToInputString : FormulaFOL -> String
ffolToInputString f =
    case f of
        Exists (Var x) g ->
            "!E[" ++ x ++ "]" ++ ffolToInputString g

        Forall (Var x) g ->
            "!A[" ++ x ++ "]" ++ ffolToInputString g

        Pred n ts ->
            n ++ termsToInputString ts

        Equal t1 t2 ->
            termToInputString t1 ++ "=" ++ termToInputString t2

        Neg g ->
            "¬" ++ ffolToInputString g

        Conj g h ->
            "(" ++ ffolToInputString g ++ "&" ++ ffolToInputString h ++ ")"

        Disj g h ->
            "(" ++ ffolToInputString g ++ "|" ++ ffolToInputString h ++ ")"

        Impl g h ->
            "(" ++ ffolToInputString g ++ "->" ++ ffolToInputString h ++ ")"

        Equi g h ->
            "(" ++ ffolToInputString g ++ "<->" ++ ffolToInputString h ++ ")"

        _ ->
            "_|_"



-- PARSER BUIDER FUCTIONS
-- It defines the syntax of a propositional variable that can be subscripting or not


folTermIdentifierParser : Parser String
folTermIdentifierParser =
    Parser.oneOf
        [ Parser.succeed ()
            |. Parser.backtrackable folTermIdentifierSubindexedParser
        , Parser.succeed ()
            |. Parser.chompIf Char.isLower
            |. Parser.chompWhile Char.isAlpha
        ]
        |> Parser.getChompedString


folTermIdentifierSubindexedParser : Parser String
folTermIdentifierSubindexedParser =
    Parser.succeed ()
        |. Parser.chompIf Char.isLower
        |. Parser.chompWhile Char.isAlpha
        |. Parser.sequence
            { start = "_{"
            , separator = ","
            , end = "}"
            , spaces = Parser.spaces
            , item = Parser.int
            , trailing = Forbidden
            }
        |> Parser.getChompedString


folPredIdentifierParser : Parser String
folPredIdentifierParser =
    Parser.oneOf
        [ Parser.succeed ()
            |. Parser.backtrackable folPredIdentifierSubindexedParser
        , Parser.succeed ()
            |. Parser.chompIf Char.isUpper
            |. Parser.chompWhile Char.isAlpha
        ]
        |> Parser.getChompedString


folPredIdentifierSubindexedParser : Parser String
folPredIdentifierSubindexedParser =
    Parser.succeed ()
        |. Parser.chompIf Char.isUpper
        |. Parser.chompWhile Char.isAlpha
        |. Parser.sequence
            { start = "_{"
            , separator = ","
            , end = "}"
            , spaces = Parser.spaces
            , item = Parser.int
            , trailing = Forbidden
            }
        |> Parser.getChompedString


folTermParser : Parser Term
folTermParser =
    Parser.oneOf
        [ succeed Func
            |. Parser.symbol "_"
            |= folTermIdentifierParser
            |= folListTermParser
        , succeed Var
            |= folTermIdentifierParser
        ]


folListTermParser : Parser (List Term)
folListTermParser =
    Parser.oneOf
        [ succeed identity
            |. Parser.symbol "("
            |= Parser.lazy (\_ -> folEnumerationTermParser)
            |. Parser.symbol ")"
        , succeed []
        ]


folEnumerationTermParser : Parser (List Term)
folEnumerationTermParser =
    folTermParser |> Parser.andThen (folEnumerationTermParserAux [])


folEnumerationTermParserAux : List Term -> Term -> Parser (List Term)
folEnumerationTermParserAux ts t =
    Parser.oneOf
        [ Parser.succeed identity
            |. Parser.symbol ","
            |= folTermParser
            |> Parser.andThen (\newt -> folEnumerationTermParserAux (ts ++ [ newt ]) t)
        , Parser.lazy (\_ -> Parser.succeed (t :: ts))
        ]



-- It defines the syntax of formulas and its processing enusuring that it corresponds to the parser of all of the string given.


ffolParser : Parser FormulaFOL
ffolParser =
    Parser.succeed identity
        |= ffolParserAux
        |. Parser.end


ffolParserAux : Parser FormulaFOL
ffolParserAux =
    Parser.oneOf
        [ Parser.succeed Exists
            |. Parser.symbol "!E"
            |. Parser.symbol "["
            |= Parser.map Var folTermIdentifierParser
            |. Parser.symbol "]"
            |= Parser.lazy (\_ -> ffolParserAux)
        , Parser.succeed Forall
            |. Parser.symbol "!A"
            |. Parser.symbol "["
            |= Parser.map Var folTermIdentifierParser
            |. Parser.symbol "]"
            |= Parser.lazy (\_ -> ffolParserAux)
        , Parser.succeed Pred
            |= folPredIdentifierParser
            |= folListTermParser
        , Parser.succeed Equal
            |= folTermParser
            |. Parser.symbol "="
            |= folTermParser
        , Parser.succeed Neg
            |. Parser.oneOf
                [ Parser.symbol "¬"
                , Parser.symbol "-"
                ]
            |= Parser.lazy (\_ -> ffolParserAux)
        , succeed identity
            |. Parser.symbol "("
            |= Parser.lazy (\_ -> expression)
            |. Parser.symbol ")"
        ]



-- It defines the propositional infixed operators (and, or, implication and equivalence)


type Operator
    = AndOp
    | OrOp
    | ImplOp
    | EquivOp


operator : Parser Operator
operator =
    Parser.oneOf
        [ Parser.succeed AndOp
            |. Parser.symbol "&"
        , Parser.succeed OrOp
            |. Parser.symbol "|"
        , Parser.succeed ImplOp
            |. Parser.symbol "->"
        , Parser.succeed EquivOp
            |. Parser.symbol "<->"
        ]


expression : Parser FormulaFOL
expression =
    ffolParserAux |> Parser.andThen (expressionAux [])



-- -- It defines the propositional infixed operators (and, or, implication and equivalence)


expressionAux : List ( FormulaFOL, Operator ) -> FormulaFOL -> Parser FormulaFOL
expressionAux revOps expr =
    Parser.oneOf
        [ Parser.succeed Tuple.pair
            |= operator
            |= ffolParserAux
            |> Parser.andThen (\( op, newExpr ) -> expressionAux (( expr, op ) :: revOps) newExpr)
        , Parser.lazy (\_ -> Parser.succeed (finalize revOps expr))
        ]



-- It defines the priority and the processing of the operators and formulas.


finalize : List ( FormulaFOL, Operator ) -> FormulaFOL -> FormulaFOL
finalize revOps finalExpr =
    case revOps of
        [] ->
            finalExpr

        -- AND EXPRESSIONS CASES
        -- And operation have the maximum priorty, so module have a unique case
        ( expr, AndOp ) :: otherRevOps ->
            finalize otherRevOps (Conj expr finalExpr)

        -- OR EXPRESSIONS CASES
        -- Or have the second maximum priority, so we need to determine how parser's going to do if it searches an and after, and if it searches something different.
        ( expr, OrOp ) :: ( expr2, AndOp ) :: otherRevOps ->
            Disj (finalize (( expr2, AndOp ) :: otherRevOps) expr) finalExpr

        ( expr, OrOp ) :: otherRevOps ->
            finalize otherRevOps (Disj expr finalExpr)

        -- IMPLICATION EXPRESSIONS CASES
        ( expr, ImplOp ) :: ( expr2, AndOp ) :: otherRevOps ->
            Impl (finalize (( expr2, AndOp ) :: otherRevOps) expr) finalExpr

        ( expr, ImplOp ) :: ( expr2, OrOp ) :: otherRevOps ->
            Impl (finalize (( expr2, OrOp ) :: otherRevOps) expr) finalExpr

        ( expr, ImplOp ) :: otherRevOps ->
            finalize otherRevOps (Impl expr finalExpr)

        -- EQUIVALATION EXPRESSIONS CASES
        ( expr, EquivOp ) :: ( expr2, AndOp ) :: otherRevOps ->
            Equi (finalize (( expr2, AndOp ) :: otherRevOps) expr) finalExpr

        ( expr, EquivOp ) :: ( expr2, OrOp ) :: otherRevOps ->
            Equi (finalize (( expr2, OrOp ) :: otherRevOps) expr) finalExpr

        ( expr, EquivOp ) :: ( expr2, ImplOp ) :: otherRevOps ->
            Equi (finalize (( expr2, ImplOp ) :: otherRevOps) expr) finalExpr

        ( expr, EquivOp ) :: otherRevOps ->
            finalize otherRevOps (Equi expr finalExpr)


{-| It returns a tuple with may be a substitution (if it can be read it), the input considerated to parse and a message of error if it is not able to performs the parsing.
-- Messages are not perfect but we're working to improve it.
-}
substitutionReadFromString : String -> ( Maybe Substitution, String, String )
substitutionReadFromString x =
    case cleanSpaces x of
        "" ->
            ( Maybe.Nothing, "", "Argument is empty" )

        s ->
            case Parser.run substitutionParser s of
                Ok y ->
                    ( Maybe.Just y, substitutionToInputString y, "" )

                Err y ->
                    ( Maybe.Nothing, s, "Error: " ++ String.replace "\"" "'" (Debug.toString y) )


{-| It generates the corresponding string input of a substitution
-}
substitutionToInputString : Substitution -> String
substitutionToInputString s =
    "{" ++ (String.join ", " <| List.map (\( x, t ) -> x ++ "/" ++ termToInputString t) <| Dict.toList s) ++ "}"


{-| It extract the formula readed. If it is Nothing it returns Insat
-}
substitutionReadExtraction : ( Maybe Substitution, String, String ) -> Substitution
substitutionReadExtraction ( s, _, _ ) =
    Maybe.withDefault Dict.empty s



-- It gives the definition of the parser to read substitutions


substitutionParser : Parser Substitution
substitutionParser =
    succeed Dict.fromList
        |= Parser.sequence
            { start = "{"
            , separator = ","
            , end = "}"
            , spaces = Parser.spaces
            , item = substitutionItemParser
            , trailing = Forbidden -- demand a trailing semi-colon
            }



-- It gives the definition to parsing an item of the subtitution


substitutionItemParser : Parser ( String, Term )
substitutionItemParser =
    succeed Tuple.pair
        |= folTermIdentifierParser
        |. Parser.symbol "/"
        |= folTermParser



--================--
-- REPRESENTATION --
--================--
-- It generates the string of a term


termToString : Term -> String
termToString t =
    case t of
        Var x ->
            x

        Func n ts ->
            n ++ termsToString ts



-- It generates the string of a list of terms (parameters)


termsToString : List Term -> String
termsToString ts =
    if List.isEmpty ts then
        ""

    else
        "(" ++ (String.join ", " <| List.map termToString ts) ++ ")"


{-| It generates the string of a First Order Logic Formula,
-}
ffolToString : FormulaFOL -> String
ffolToString f =
    case f of
        Pred n ts ->
            n ++ termsToString ts

        Equal t1 t2 ->
            termToString t1 ++ " = " ++ termToString t2

        Neg p ->
            "¬ " ++ ffolToString p

        Conj p q ->
            "( " ++ ffolToString p ++ " ∧ " ++ ffolToString q ++ " )"

        Disj p q ->
            "( " ++ ffolToString p ++ " ∨ " ++ ffolToString q ++ " )"

        Impl p q ->
            "( " ++ ffolToString p ++ " → " ++ ffolToString q ++ " )"

        Equi p q ->
            "( " ++ ffolToString p ++ " ↔ " ++ ffolToString q ++ " )"

        Exists (Var v) p ->
            "∃" ++ v ++ " " ++ ffolToString p

        Forall (Var v) p ->
            "∀" ++ v ++ " " ++ ffolToString p

        _ ->
            "⊥"


{-| It generates the string of a First Order Logic Set of formulas,
-}
sfolToString : SetFOL -> String
sfolToString fs =
    "{" ++ (String.join ", " <| List.map ffolToString fs) ++ "}"



-- It generates the string of a term using latex notation


termToMathString : Term -> String
termToMathString t =
    case t of
        Var x ->
            x

        Func n ts ->
            n ++ termsToMathString ts



-- It generates the string of a list of terms (parameters) using latex notation


termsToMathString : List Term -> String
termsToMathString ts =
    if List.isEmpty ts then
        ""

    else
        "\\left( " ++ (String.join ", " <| List.map termToMathString ts) ++ "\\right)"


{-| It generates the string of a First Order Logic Formula using latex notation
-}
ffolToMathString : FormulaFOL -> String
ffolToMathString f =
    case f of
        Pred n ts ->
            n ++ termsToMathString ts

        Equal t1 t2 ->
            termToMathString t1 ++ " = " ++ termToMathString t2

        Neg p ->
            "¬ " ++ ffolToMathString p

        Conj p q ->
            "( " ++ ffolToMathString p ++ " ∧ " ++ ffolToMathString q ++ " )"

        Disj p q ->
            "( " ++ ffolToMathString p ++ " ∨ " ++ ffolToMathString q ++ " )"

        Impl p q ->
            "( " ++ ffolToMathString p ++ " → " ++ ffolToMathString q ++ " )"

        Equi p q ->
            "( " ++ ffolToMathString p ++ " ↔ " ++ ffolToMathString q ++ " )"

        Exists (Var v) p ->
            "∃" ++ v ++ " " ++ ffolToMathString p

        Forall (Var v) p ->
            "∀" ++ v ++ " " ++ ffolToMathString p

        _ ->
            "⊥"


{-| It generates the string of a First Order Logic Set of formulas using latex notation
-}
sfolToMathString : SetFOL -> String
sfolToMathString fs =
    "\\begin{array}{l} " ++ (String.join " \\\\ " <| List.map ffolToMathString fs) ++ " \\end{array}"


{-| It gives a string representation of a form tree
-}
formTreeToString : Graph FormulaFOL () -> String
formTreeToString t =
    Graph.toString (\x -> Just <| ffolToString x) (\_ -> Nothing) t


{-| It gives a string representation of a form tree using DOT format
-}
formTreeToDOT : Graph FormulaFOL () -> String
formTreeToDOT t =
    let
        myStyles =
            { defaultStyles | node = "shape=plaintext, color=black", edge = "dir=none" }
    in
    Graph.DOT.outputWithStyles myStyles (\x -> Just <| ffolToString x) (\_ -> Nothing) t


{-| It gives a string representation of a substitution using latex notation
-}
substitutionToString : Substitution -> String
substitutionToString s =
    "{" ++ (String.join ", " <| List.map (\( x, t ) -> x ++ "/" ++ termToString t) <| Dict.toList s) ++ "}"


{-| It gives a string representation of a substitution using latex notation
-}
substitutionToMathString : Substitution -> String
substitutionToMathString s =
    "\\left\\lbrace" ++ (String.join ", " <| List.map (\( x, t ) -> x ++ "/" ++ termToMathString t) <| Dict.toList s) ++ "\\right\\rbrace"


{-| It gives a L-structure as string. It needs a funtion that gives the string representation of the elements of the L-Structure
-}
l_StructureToString : L_Structure a -> (a -> String) -> String
l_StructureToString ( u, i ) ra =
    let
        universeString =
            "U: {" ++ (String.join ", " <| List.map ra u) ++ "}"

        constString =
            "c={" ++ (String.join ", " <| List.map ra <| Dict.values i.const) ++ "}"

        funcString =
            let
                fInternalString f ar f_M =
                    String.join ";" <|
                        List.map
                            (\e -> f ++ "(" ++ (String.join "," <| List.map ra e) ++ ")=" ++ (Maybe.withDefault "" <| Maybe.map ra (f_M e)))
                            (LE.cartesianProduct <| List.repeat ar u)
            in
            let
                fString ( f, ( ar, f_M ) ) =
                    f ++ ": {\n" ++ fInternalString f ar f_M ++ "\n}"
            in
            String.join "\n" <| List.map fString <| Dict.toList i.func

        predString =
            let
                pInternalString ar p_M =
                    String.join ";" <|
                        List.foldr
                            (\e ac ->
                                if Maybe.withDefault False <| p_M e then
                                    ("(" ++ (String.join "," <| List.map ra e) ++ ")") :: ac

                                else
                                    ac
                            )
                            []
                            (LE.cartesianProduct <| List.repeat ar u)
            in
            let
                pString ( p, ( ar, p_M ) ) =
                    p ++ ": {\n" ++ pInternalString ar p_M ++ "\n}"
            in
            String.join "\n" <| List.map pString <| Dict.toList i.pred
    in
    String.join "\n\n" [ universeString, constString, funcString, predString ]
